#!/usr/bin/env python
# cspell:ignore Bambu Fiberon
"""Container for slicer presets.

See slicer_tools.py for more comprehensive documentation.
"""

import json

from copy import deepcopy
from pathlib import Path
from sys import platform
from typing import cast, NamedTuple, Generator

from slicer_tools.common import AllNodeSettings, NodeMetadata, PresetGroup, PresetType
from slicer_tools.common import SettingsDict
from slicer_tools.common import FROM, INHERITS, NAME, DEFAULT_ENCODING


class PresetPath(NamedTuple):
    """Provide a unique path/key for presets in ProjectPresets.

    This tuple eliminates the requirement sparse nested dictionaries or minimal classes
    emulating nested dicts (at the cost of using tuple hashes for dict keys).
    """

    # Note that this path should be guaranteed to be unique within a project (.3mf)
    # file, but is NOT guaranteed unique across files. See DiffValuePath to deal with
    # this problem.
    type: PresetType
    preset_name: str


class PresetNode:
    """Container for preset or override definition.

    A preset DataNode contains all of the information required for preset/override
    definition.

    The primary data managed by this class is "the collection of key/value pairs for any
    settings" that differ from "key/value pairs the preset inherits from its immediate
    parent". That is, a Preset effectively represents the difference in settings between
    the source preset and the parent of the source preset.

    A Preset does not address any differences further up the inheritance tree - these
    are generated by walking the inheritance tree (addressed in PresetNodes container
    class.

    A PresetNode contains settings for single preset type, and may be sourced from a
    json preset file, a preset config file in a 3mf archive, or a preset override from a
    project config file in a 3mf archive.

    Intended as a read-only source for settings key/value pairs and other identity
    parameters.
    - filename is the name of the source file, which may be a normal file or a file
    in a 3mf archive.
    - name is the name of the preset. Note that ProjectPresets (a container class for
    Presets) requires that each name in a 3mf project is unique. See ThreeMFPresets for
    a sample implementation to manage this.
    - path is the path to the source preset file if the source is a normal file, None
    if the source is a 3mf archive member.
    """

    metadata: NodeMetadata
    path: Path | None = None
    # Settings is a private variable to support lazy loading of system and user
    # preset settings.
    _settings: SettingsDict | None

    # This is a REAL hack. BambuStudio more or less treats configs in
    # base directories in user folders as more-or-less system configs.
    # We use this flag to modify these on the fly to more-or-less follow
    # the way BS behaves.
    _user_base_fix: bool = False

    def __init__(
        self,
        metadata: NodeMetadata,
        settings: SettingsDict | None = None,
        path: Path | None = None,
        user_base_fix: bool = False,
    ) -> None:
        """Create preset instance."""
        # PresetNode is designed around lazy loading system and user settings.
        # If the default value/None is provided for settings, lazy loading will
        # be assumed.
        # self._settings is initialised here, and lazy loading occurs in the data
        # getter.
        self._settings = settings
        self.path = path
        if user_base_fix:
            self._user_base_fix = user_base_fix
            # Faux system group in the base folder. Second hack below.
            self.metadata = metadata._replace(group=PresetGroup.SYSTEM)
        else:
            self.metadata = metadata

        if self._settings is None and self.path is None:
            # Shouldn't happen, but in case.
            raise ValueError(
                "Settings and path arguments cannot both be None for a PresetNode()"
            )

    @property
    def settings(self) -> SettingsDict:
        """Return the node settings dictionary."""
        # see __init__ for an explanation of _settings and the settings getter.
        if self._settings is None:
            if self.path is None:
                # I think this should be caught during initialisation. But belt and
                # braces.
                raise KeyError(
                    "Preset node contains no settings data and no path to preset file."
                )

            # Lazy load needed for settings.
            with open(self.path, encoding=DEFAULT_ENCODING) as fp:
                self._settings = json.load(fp)
                if self._user_base_fix:
                    # This is our second hack to make user/**/base/*.json look
                    # like system presets.
                    self._settings[FROM] = PresetGroup.SYSTEM

        # Relying on the caller to not modify settings.
        return self._settings


class ProjectPresets:
    """Container for a group of related Preset nodes built around a BS .3mf model.

    ProjectPresets provides access to a related group of presets and overrides, both at
    the level of nodes in the inheritance tree, and as roll ups of the inheritance
    tree for a node. As the main preset grouping I'm working with is a BambuStudio
    .3mf project file, so this class is based around a .3mf structure.

    The container manages two groups of Preset nodes:
    - Shared preset nodes available to all instances of ProjectSettings. These are:
        - BambuStudio system presets.
        - BambuStudio user presets for one user, which defaults to the last user to
        login to BambuStudio.
        - Note the shared presets will be a superset of the system/user presets used by
        any given project file, but it's convenient to grab them all at once.
    - A set of per instance project preset nodes, corresponding to:
        - Project (3mf file unit) settings for filament, machine and process presets.
        - Project overrides of any of the above presets.
    The shared settings load on the first instantiation of ProjectSettings, and can only
    be overridden on this first call.
    """

    # Both preset node dictionaries are keyed by PresetPath named tuples.
    # Should be enough for relatively efficient returns, but DOES require creation of
    # unique names for project overrides of presets.
    _project_nodes: dict[PresetPath, PresetNode]
    # Class variable, share among all instances.
    _shared_nodes: dict[PresetPath, PresetNode] = {}

    def __init__(self, appdata_path: Path | None = None, bbl_user_id: str = "") -> None:
        r"""Create node containers and load system and user preset nodes.

        appdata_path and bbl_user_id will only be used in the first instance call, and
        ignored thereafter.

        appdata_path points to the location for the Bambu Studio system snd user
        preset nodes, and is the source for _shared_settings.
        If not specified, this defaults to
            Windows: c:\users\<username>\appdata\Roaming\BambuStudio
            macos: /Users/user.name/Library/Application Support/BambuStudio
        I don't have a mac, so haven't tested the macos version.

        If specified, the folder structure must follow the form:
            appdata_path
                BambuStudio.conf (only required if bbl_user_id is not specified)
                system
                    BBL.json (must follow the same format as the default version)
                    BBL
                        filament
                        machine
                        process
                user
                    <bbl_user_id>
                        filament
                            base
                        machine
                            base
                        process
                            base
        bbl_user_id is the user id of the folder containing user presets. If it is not
        specified, a default value will be extracted from BambuStudio.conf (this will
        typically be the id of the last user to login to BS). Initialisation will fail
        dramatically if user_id is not specified and BambuStudio.conf does not exist.
        """
        if len(ProjectPresets._shared_nodes) == 0:
            # First instance of this class, shared settings don't exist.
            self._load_shared_nodes(appdata_path, bbl_user_id)

        # Create instance project dict.
        self._project_nodes = {}

    @classmethod
    def _load_shared_nodes(
        cls, appdata_path: Path | None = None, bbl_user_id: str = ""
    ) -> None:
        """Load shared nodes dictionary."""
        # Given this is the only place these paths should appear, happy enough hard
        # coding here.
        studio_path = appdata_path
        if studio_path is None:
            if platform == "win32":
                studio_path = Path.home() / "Appdata/Roaming"
            elif platform == "darwin":
                studio_path = Path.home() / "Library/Application Support"
            else:
                raise RuntimeError(
                    f"BambuScan doesn't know how to find BambuStudio"
                    f" folder on {platform}"
                )
            studio_path = studio_path / "BambuStudio"

        user_id = bbl_user_id
        if not user_id:
            try:
                with open(
                    studio_path / "BambuStudio.conf", encoding=DEFAULT_ENCODING
                ) as fp:
                    # strip out checksum comment. Hells bells.
                    raw_json = "".join(line for line in fp if not line.startswith("#"))
            except FileNotFoundError:
                print("Missing BambuStudio.conf file.")
                raise

            json_data = json.loads(raw_json)
            try:
                user_id = json_data["app"]["preset_folder"]
            except KeyError:
                print("Missing 'preset folder' key/value in BambuStudio.conf")
                raise

        # Populate system presets.
        cls._system_nodes_from_bbl_json(studio_path / "system")

        # Walk paths. Should just verify system, loads user.
        # Duplicates will get a printed warning, but no error.
        cls._nodes_from_path_walk(
            root=studio_path / "system/BBL", group=PresetGroup.SYSTEM
        )
        cls._nodes_from_path_walk(
            root=studio_path / "user" / user_id, group=PresetGroup.USER
        )

        # And just in case we didn't find anything, add a dummy entry so
        # we don't go through this on each load.
        # Likely the caller will find out the hard way.
        if len(cls._shared_nodes) == 0:
            cls._shared_nodes[PresetPath(PresetType.FILAMENT, "dummy")] = PresetNode(
                NodeMetadata(
                    name="Dummy",
                    filename="Dummy.json",
                    group=PresetGroup.SYSTEM,
                    preset_type=PresetType.PROCESS,
                ),
                settings={"name": "No standard settings found"},
            )

    @classmethod
    def _system_nodes_from_bbl_json(cls, system_path: Path) -> None:
        """Load system preset file locations from BBL.json.

        BBL.json is assumed definitive. I know there are duplicates. E.g. Fiberon.
        We will warn about these and ignore the duplicates.
        """
        # Some hard coding of strings that should only occur here.
        with open(system_path / "BBL.json", encoding=DEFAULT_ENCODING) as fp:
            data = json.load(fp)

        for preset_type in PresetType:
            # ignoring machine models data.
            for item in data[preset_type + "_list"]:
                name = item[NAME]
                path = system_path / "BBL" / item["sub_path"]

                preset_key = PresetPath(preset_type, name)
                if preset_key in cls._shared_nodes:
                    raise ValueError(
                        # Warn, but don't do anything about it.
                        f"Highly unexpected duplicate definition in BBL.json:"
                        f"   '{preset_key}'."
                    )

                # Prep for lazy load.
                cls._shared_nodes[preset_key] = PresetNode(
                    NodeMetadata(
                        name=name,
                        filename=path.name,
                        group=PresetGroup.SYSTEM,
                        preset_type=preset_type,
                    ),
                    path=path,
                )

    @classmethod
    def _nodes_from_path_walk(cls, root: Path, group: PresetGroup) -> None:
        """Load preset nodes from path walk. Really for user paths."""
        for preset_type in PresetType:
            base_path = root / preset_type
            for json_path in base_path.glob("**/*.json"):
                preset_key = PresetPath(
                    type=preset_type,
                    preset_name=json_path.stem,
                )
                if preset_key in cls._shared_nodes:
                    if json_path != cls._shared_nodes[preset_key].path:
                        # Warn but don't do anything about it.
                        print(f"Warning duplicate definition for '{preset_key}`.")

                else:
                    # Tell PresetNode to deal with user bases that are actually
                    # system types. (Handled by Preset Node as multiple fixes required.)
                    base_fix = False
                    if group == PresetGroup.USER and json_path.parent.name == "base":
                        base_fix = True

                    cls._shared_nodes[preset_key] = PresetNode(
                        NodeMetadata(
                            name=json_path.stem,
                            filename=json_path.name,
                            group=group,
                            preset_type=preset_type,
                        ),
                        path=json_path,
                        user_base_fix=base_fix,
                    )

    def add_project_node(
        self,
        metadata: NodeMetadata,
        settings: SettingsDict,
    ) -> None:
        """Add project preset node based on the data dict."""
        key = PresetPath(
            type=metadata.preset_type,
            preset_name=metadata.name,
        )

        # Quick and dirty checks on uniqueness.
        if key in self._shared_nodes:
            raise KeyError(
                f"Attempt to add project preset {key} with the same name as a system"
                f" preset."
            )
        if key in self._project_nodes:
            raise KeyError(
                f"Attempt to add project preset {key} when preset already exists."
            )
        # Arguably should check name is unique, but leaving that up to BS for now.
        # preset_type = settings[]
        self._project_nodes[key] = PresetNode(
            metadata=metadata,
            settings=settings,
        )

    def _node(self, preset_type: PresetType, name: str) -> PresetNode:
        """Return the specified preset node."""
        key = PresetPath(type=preset_type, preset_name=name)
        if key in self._shared_nodes:
            return self._shared_nodes[key]

        if key in self._project_nodes:
            return self._project_nodes[key]

        raise KeyError(f"Undefined preset '{key}'.")

    def node_settings(self, preset_type: PresetType, name: str) -> SettingsDict:
        """Return a **deep copy** of the node settings named preset node.

        The settings returned by this call are the differences between the settings
        for the named node and the settings of the parent node (i.e. the preset data
        that would appear in the json file for the preset).
        """
        # Keep the source data read only.
        return deepcopy(self._node(preset_type, name=name).settings)

    def all_node_settings(
        self,
        preset_type: PresetType,
        node_name: str,
        ref_node: str = "",
        ref_group: PresetGroup | None = None,
    ) -> AllNodeSettings:
        """Return all settings for node_name of preset_type.

        Settings may be provided as either
        - a single SettingsDict and associated NodeMetadata for node_name (default), or
        - as partitioned settings consisting of a SettingsDict generated by walking the
        inheritance from node_name up to but not including the reference node, and a
        reference settings dictionary from walking the inheritance tree from the
        reference node to the root of the tree. (i.e. reference settings and
        node_name settings that are different to the reference settings).

        The reference node is determined from either the reference node name ref_node
        or the youngest inherited ancestor of node_name that has a PresetGroup of
        ref_group. At least one of ref_value and ref_group must be empty ("").

        If called with default settings, the return structure is:
        - ret_val.metadata contains metadata for node_name.
        - ret_val.settings contains all settings for node_name.
        - ret_val.ref_metadata = None.
        - ret_val.reference = {}

        If called with with a reference group/node specified, the return structure is:
        - ret_val.metadata contains metadata for node_name.
        - ret_val.settings contains all settings different to the reference settings.
        - ret_val.ref_metadata contains the metadata for the reference node.
        - ret_val.reference contains all settings for the reference node.
        """
        if ref_group and ref_node:
            raise ValueError(
                f"Cannot specify both reference group '{ref_group}'"
                f" and reference node '{ref_node}'"
            )

        if ref_group:
            # A bit messy - the preset may not inherit from the ref_group, but if so,
            # should inherit from its parents. So we build a check set for all possible
            # reference groups.
            include = False
            ref_set = []
            for group in PresetGroup:
                if ref_group == group:
                    include = True
                if include:
                    ref_set.append(group)

        # Unfortunately because of the way python dictionary unions work (keep
        # rightmost value), we need to walk the tree first and then build the
        # return dictionaries with union assignment |= from the root back to the
        # calling node. I've opted to keep this in a a single method, as I think the
        # logic is cleaner than via recursion.
        # First up create node lists for the settings/different to reference settings
        # and the reference settings.
        settings_nodes: list[SettingsDict] = []
        reference_nodes: list[SettingsDict] = []
        # Collect settings until we find the first reference node and then
        # switch.
        target = settings_nodes

        # Current node for processing.
        this_node: PresetNode | None = self._node(preset_type, node_name)
        # metadata for this node.
        metadata: NodeMetadata = cast(PresetNode, this_node).metadata
        ref_metadata: NodeMetadata | None = None

        while this_node:
            # This next bit looks stupid, but I think it's correct.
            # The first thing we do is check if we need to switch to gathering reference
            # nodes. This may have the odd side effect that the original node may
            # be assigned directly to the reference settings if it is in the reference
            # group.
            if not ref_metadata and (
                (ref_group and this_node.metadata.group in ref_set)
                or (this_node.metadata.name == ref_node)
            ):
                # This is either the youngest ancestor in reference group, or this is
                # the reference node, so we switch to gathering reference data.
                ref_metadata = this_node.metadata
                target = reference_nodes

            # Warning - we are grabbing mutable node SettingsDicts here.
            # Caller should not alter these.
            target.append(this_node.settings)

            # Move on to parent node.
            # By rights inherits should ALWAYS be a str, but sometimes doesn't exist at
            # all.
            if INHERITS not in this_node.settings:
                inherits = ""
            else:
                inherits = cast(str, this_node.settings[INHERITS])

            if inherits:
                this_node = self._node(preset_type, inherits)
            else:
                this_node = None

        # Prepare empty settings dicts and
        # unroll the lists to populate the dicts, and finally sort them.
        working: SettingsDict = {}
        for node in reversed(settings_nodes):
            working |= node
        settings = dict(sorted(working.items()))
        working = {}
        for node in reversed(reference_nodes):
            working |= node
        reference = dict(sorted(working.items()))

        # I don't think this is possible, but no harm in a quick check.
        if len(reference) != 0 and not ref_metadata:
            raise ValueError(
                "Reference values defined, but reference metadata is not."
                "\nThis should not be possible."
            )

        return AllNodeSettings(
            metadata=metadata,
            ref_metadata=ref_metadata,
            source_subtree=settings,
            reference_subtree=reference,
        )

    def project_presets(
        self, reference_group: PresetGroup | None = None, reference_node: str = ""
    ) -> Generator[AllNodeSettings]:
        """Yield settings and metadata for project presets relative to reference group.

        The generator repeatedly calls all_node_settings to generate and return
        the setting for each project preset node in the ProjectPresets instance.

        If reference group or reference_node is specified, these settings are all
        relative to the reference node/group (see all_node_settings for more detail),
        which is also returned in AllNodeSettings.
        """
        for node in self._project_nodes.values():
            # Dom't need keys as we iterate over all elements in _project_nodes,
            # and key data is repeated in node.metadata.
            yield self.all_node_settings(
                node.metadata.preset_type,
                node_name=node.metadata.name,
                ref_group=reference_group,
                ref_node=reference_node,
            )

    def preset_names(self, preset_type: PresetType) -> tuple[str]:
        """Return the list of names of the project presets of the specified type."""
        name_list: tuple[str] = cast(tuple[str], tuple(
            # pylint: disable-next=consider-iterating-dictionary
            x.preset_name for x in self._project_nodes.keys() if x.type == preset_type
        ))

        return name_list



#!/usr/bin/env python
# cspell:ignore Bambu Fiberon Prusa
# TODO: Add special types to formatting (name, inherits, etc.) - also make version
# xx.xx.xx.xx. Document all of this.
# TODO: INCLUDE HUGE WARNING ABOUT INHERITING FROM USER DEFINED BASE TYPES (I
# create these as system types - they are very much not - see _user_base_fix).
# TODO: Document that differencing is only reliable for system group. Others are
# untested.
"""Quick and dirty  tools for reading and comparing Bambu Studio slicer presets.

Summary: This package includes classes for loading system and user presets, and
presets/overrides specified in model .3mf files. The main feature of the package is
comparing presets in one or more .3mf files (the default mode summarises differences
relative to system presets). Differences are provided in .xlsx format. In addition, all
settings in a single preset can be dumped as a json file (from walking the inheritance
tree).

Notes:
    - In this package, I focus on filament, machine and process as the core preset
    types.
    - Preset types may be system, user and project (these three follow the BambuStudio
    naming convention) and **override**, which is a preset generated from project file
    overrides of one of three other groups. Override is not BambuStudio terminology,
    but provides a convenient naming convention for handling project file oddities.

Cautions:
    - This implementation has only been tested on  Windows. But if the Bambu Studio file
    structure for macos mimics Windows, it should work OK on there too (or at least
    can be made to work with minimal updates).
    - I have only implemented an english version, so this package may break when using
    other languages.
    - Right now the package is built around Bambu Studio files. However, I expect that
    it should be easy to extend to Orca Slicer files, and possibly also Prusa Slicer.
"""
import json
import string
import random
from argparse import ArgumentParser, Namespace
from enum import Enum, StrEnum
from typing import cast, NamedTuple, Generator
from zipfile import ZipFile
from zipfile import Path as zPath
from pathlib import Path

from openpyxl import Workbook  # type: ignore[import-untyped]

from slicer_tools.common import AllNodeSettings, NodeMetadata, PresetType, PresetGroup
from slicer_tools.common import SettingsDict, SettingValue
from slicer_tools.common import FROM, INHERITS, NAME, DEFAULT_ENCODING
from slicer_tools.common import choose
from slicer_tools.presets import ProjectPresets

# Navigation and file name components
THREE_MF_FOLDER = "./3MF"
THREE_MF_EXT = ".3mf"
CONFIG_EXT = ".config"
METADATA = "Metadata/"

# Constants for manipulating .configs
# Can build all json file names from these two constants and the
# PresetType enum below.
SETTINGS = "_settings"
PROJECT_CONFIG_NAME = "project" + SETTINGS
EXCLUDE_CONFIGS = ["model_settings", "slice_info"]

# more json key elements
# annoyingly, in config files, settings id uses a different terminology to
# the file naming - specifically print in the json is equivalent to process and
# printer is equivalent to machine. Yay!
PRINT = "print"
PRINTER = "printer"
DIFFS_TO_SYSTEM = "different_settings_to_system"
ID = "_id"
INHERITS_GROUP = "inherits_group"  # Because why do one thing one way ...
VERSION = "version"
SPECIAL_VERSION = "XX.XX.XX.XX"
# Special settings that will require .
SPECIAL_TYPES = [
    "from",
    "filament_extruder_variant",
    "filament_settings_id",
    "inherits",
    "name",
    "print_extruder_id",
    "print_extruder_variant",
    "print_settings_id",
    "printer_extruder_id",
    "printer_extruder_variant",
    "printer_settings_id",
    "version",
]


class CellFormat(StrEnum):
    """Builtin Excel cell formats."""

    NORMAL = "Normal"
    GOOD = "Good"
    BAD = "Bad"
    INPUT = "Input"
    NOTE = "Note"
    NEUTRAL = "Neutral"
    HEADING4 = "Headline 4"


class DiffType(Enum):
    """Difference types."""

    # No difference in values.
    NO_DIFF = 0
    # Reference value (the value differenced to!).
    REFERENCE = 1
    # Normal difference to reference value.
    DIFFERENCE = 2
    # Override system difference.
    OVERRIDE = 3
    # Unset reference value.
    UNSET = 4
    # Too complex to diff (for now at least).
    COMPLEX = 5
    # Value requires special/manual handling
    SPECIAL = 6


class CellInfo(NamedTuple):
    """Summary data for writing to an Excel cell."""

    row: int
    column: int
    value: str
    format: DiffType | CellFormat


class DiffValue(NamedTuple):
    """Container for difference value and difference type information."""

    value: str
    type: DiffType


class DiffValuePath(NamedTuple):
    """Provide a unique path/key for difference sets.

    This tuple eliminates the requirement for sparse nested dictionaries or minimal
    classes emulating nested dicts, at the cost of using tuples as dict keys.
    """

    # This is pretty similar to PresetPath. However, a diff set may contain multiple
    # preset types and multiple .3mf files, so we need to add  the filename to ensure
    # uniqueness - as this info is already available in node metadata, we're just using
    # that. In addition, the value (row) name will also be sparse. As adding row name
    # to the key makes iteration over differences easier, I've done that as well.
    row_name: str
    metadata: NodeMetadata


class DiffMatrix:
    """Difference data in a sparse matrix form."""

    # This started as a dataclass, but as all members are now private, converted to
    # full fledged class.
    _rows: dict[str, int]
    _cols: dict[NodeMetadata, int]
    _values: dict[DiffValuePath, DiffValue]
    # Flag indicating indices need recalculating. Set to true every time a new
    # value is added.
    _reset_required: bool
    # Group, filename, preset name.
    _header_count: int = 3

    def __init__(self) -> None:
        """Create instance variables."""
        self._rows = {}
        self._cols = {}
        self._values = {}
        self._reset_required = True

    def column_exits(self, metadata: NodeMetadata) -> bool:
        """Test if column is defined."""
        if metadata in self._cols:
            return True
        return False

    def add_value(
        self, row_name: str, column_id: NodeMetadata, value: str, value_type: DiffType
    ) -> None:
        """Add or overwrite a value to the diff matrix.

        Automatically triggers an instance reset and will make any active generator
        invalid (values/indices have been modified in ways that are unpredictable for
        the generator).
        """
        self._reset_required = True
        self._rows[row_name] = -1
        self._cols[column_id] = -1
        self._values[DiffValuePath(row_name, column_id)] = DiffValue(value, value_type)

    def _reset_lookups(self) -> None:
        """Prepare row and column lookup indices for writing tables.

        All indices are 0 based.
        """
        key: str | NodeMetadata

        keys = self._rows.keys()
        for i, key in enumerate(sorted(keys), start=0):
            self._rows[key] = i

        # Column sort is not worth doing, as we want a rough sort along the lines:
        #   Reference, Override, Project and User.
        # While it would be possible to write a compare function, it's nearly as
        # simple to just run over the column list for each group.
        i = 0
        for group in [
            PresetGroup.SYSTEM,
            PresetGroup.OVERRIDE,
            PresetGroup.PROJECT,
            PresetGroup.USER,
        ]:
            for key, _ in self._cols.items():
                if key.group == group:
                    self._cols[key] = i
                    i += 1

    def table_cells(self) -> Generator[CellInfo]:
        """Generate table and associated row names and column headers.

        Return values are row offset (0 based), column number, value, value type or
        format.
        """
        if self._reset_required:
            self._reset_lookups()

        # Header names. All 0 based.
        yield CellInfo(0, 0, "Group", CellFormat.NORMAL)
        yield CellInfo(1, 0, "Filename", CellFormat.NORMAL)
        yield CellInfo(2, 0, "Preset", CellFormat.NORMAL)

        for row_name, row_offset in self._rows.items():
            yield CellInfo(
                row_offset + self._header_count, 0, row_name, CellFormat.NORMAL
            )

        for metadata, col_offset in self._cols.items():
            yield CellInfo(0, col_offset + 1, metadata.group.value, CellFormat.NORMAL)
            yield CellInfo(1, col_offset + 1, metadata.filename, CellFormat.NORMAL)
            yield CellInfo(2, col_offset + 1, metadata.name, CellFormat.NORMAL)

        for key, value in self._values.items():
            row = self._rows[key.row_name] + self._header_count
            col = self._cols[key.metadata] + 1
            yield CellInfo(row, col, value.value, value.type)

    def row_count(self) -> int:
        """Row count in the table include header lines."""
        return len(self._rows) + self._header_count


class ThreeMFPresets:
    """Container for all presets in a 3mf file.

    Currently skips per object and slice data. And will probably never implement these.
    """

    filename: str
    # Project settings extracted from project_settings.config.
    project_config: SettingsDict

    # Collection of the following presets:
    # - All project presets and overrides in the project file.
    # - All user presets that the project presets/overrider might inherit from.
    # - All system presets that the project presets/overrides might inherit from.
    # ThreeMFPresets creates the first group, the other two are loaded automatically.
    # After all project presets have been created, the project_presets member
    # can generate settings and differences as required.
    # The caveat is that calls to generate settings/differences must be deferred until
    # all Presets in the 3mf have been instantiated, as otherwise there is a risk that
    # the project_presets object will not hold the data required for a complete roll up.
    # (In effect this is creating a half way house to a class variable that is only
    # visible within each ThreeMFSettings instance and the Settings instances contained
    # by it. Yuck!)
    presets: ProjectPresets

    def __init__(
        self, project_file: Path, appdata_path: Path | None = None, user_id: str = ""
    ) -> None:
        """Create project metadata container.

        appdata_path and user_id override the default locations for find the system
        and user presets.
        """
        self.filename = project_file.name
        self.presets = ProjectPresets(appdata_path=appdata_path, bbl_user_id=user_id)
        with ZipFile(project_file, mode="r") as archive:
            for zip_path in zPath(archive, at=METADATA).iterdir():
                # Iterate over metadata folder to find presets (.config files).
                if (
                    zip_path.name.endswith(CONFIG_EXT)
                    and zip_path.stem not in EXCLUDE_CONFIGS
                ):
                    # Excluding xml format configs for now.
                    # We'll need the config data no matter what.
                    # By rights, this should be a json file at this point.
                    settings = json.loads(zip_path.read_text(encoding=DEFAULT_ENCODING))
                    if zip_path.stem == PROJECT_CONFIG_NAME:
                        # Project config. Will contain multiple overrides and
                        # requires special handling.
                        self.project_config = settings
                        self._process_project_settings()
                    else:
                        # Find out what type we are working with and create preset.
                        for prefix in PresetType:
                            if zip_path.stem.startswith(prefix + SETTINGS):
                                self.presets.add_project_node(
                                    NodeMetadata(
                                        name=settings[NAME],
                                        filename=project_file.name,
                                        preset_type=prefix,
                                        group=PresetGroup.PROJECT,
                                    ),
                                    settings=settings,
                                )
                                break

    def _process_project_settings(self) -> None:
        """Extract presets from project_settings.config and add to project presets."""
        # process the differences by type.
        # Guaranteed there is a better way to this, but patience is gone.
        group = PresetGroup.OVERRIDE
        filament_count = len(self.project_config[PresetType.FILAMENT + SETTINGS + ID])
        filament_idx = -1
        for diff_idx, diff in enumerate(self.project_config[DIFFS_TO_SYSTEM]):
            if diff_idx == 0:
                # Process.
                name = self.project_config[PRINT + SETTINGS + ID]
                value_idx = 0
                preset_type = PresetType.PROCESS

            elif diff_idx == filament_count + 1:
                # Machine.
                name = self.project_config[PRINTER + SETTINGS + ID]
                value_idx = 0
                preset_type = PresetType.MACHINE

            else:
                # Filament
                filament_idx += 1
                name = self.project_config[PresetType.FILAMENT + SETTINGS + ID][
                    filament_idx
                ]
                value_idx = filament_idx
                preset_type = PresetType.FILAMENT

            # Set up the special unicorn tears preset.
            # Create settings and add key/values missing from the override differences.
            # Unfortunately, the project override name is not unique, as it is also the
            # the name of the immediate parent that the override inherits from (see
            # PresetMetadata for more on this, and why we are setting up the fix we
            # do here).
            # The fix is to create the preset metadata with the parent preset inherit
            # specified and the preset name made unique by adding 4 random characters
            # to the end of the name.
            name = cast(str, name)
            metadata = NodeMetadata(
                name=name + "_" + "".join(random.sample(string.ascii_lowercase, 4)),
                filename=self.filename,
                preset_type=preset_type,
                group=group,
                # This is the immediate parent, which may be different from the
                # inherits_group value in project_settings. Yay!
                override_inherits=name,
            )

            # And now we pre-populate settings with values that project_settings does
            # not provide. As another project_settings wrinkle: if the project does not
            # override the system values at all, "inherits" is set to "".
            # We do some munging here to fix this as well. Yay.
            inherits = self.project_config[INHERITS_GROUP][diff_idx]
            if not inherits:
                inherits = name

            settings = {
                # Use unique name for override.
                NAME: metadata.name,
                INHERITS: inherits,
                FROM: self.project_config[FROM],
                # special value here, as versions can be all over the place.
                VERSION: SPECIAL_VERSION,
            }

            # Finally collect the actual differences.
            self._differences_to_settings(
                value_idx=value_idx,
                filament_count=filament_count,
                diff=diff,
                settings=settings,
                metadata=metadata,
            )

            self.presets.add_project_node(
                metadata=metadata,
                settings=settings,
            )

    def _differences_to_settings(
        self,
        value_idx: int,
        diff: str,
        settings: SettingsDict,
        metadata: NodeMetadata,
        filament_count: int = 0,
    ) -> None:
        """Create override settings from difference list.

        expect_count is the number of values expected in the value lists. Used for
        list length validations. List length validation tries to be smart, but may get
        things wrong. Check errors and warnings to resolve.
        """
        diff_keys = diff.split(";")
        if not diff_keys[0]:
            return

        # If I was confident about value counts, could do this with a comprehension.
        # But instead we'll do it slow with checks. And we need to, because of
        # inconsistency.
        for key in diff_keys:
            values = self.project_config[key]

            # Deal with filament first.
            if metadata.preset_type == PresetType.FILAMENT:
                if key == "filament_notes":
                    print(
                        f"Warning: key `filament_notes' appears in override"
                        f" {metadata.name}. Key ignored."
                        f"\n  (BambuStudio currently retains the note value for one"
                        f"  filament only in the config file and I have no idea which"
                        f"  one is important.)"
                    )
                    continue

                if not isinstance(values, list):
                    raise TypeError(
                        f"Expected list of values for setting '{key}'"
                        f"\n  of filament override {metadata.name}."
                        f"\n  Got {type(values)}. (Values = {values})."
                    )

                if len(values) != filament_count:
                    # Will raise a warning or error. Prep info string accordingly.
                    raise IndexError(
                        f"Unexpected override list size."
                        f"\n  List length is expected to match override/project"
                        f" filament count."
                        f"\n  Override `{metadata.name}` setting `{key}`:"
                        f"\n     Expected list length count {filament_count}, got"
                        f" {len(values)}."
                        f"\n    (Values = {values})"
                    )

                # If we're here, I think it's real.
                settings[key] = values[value_idx]

            else:
                # Now dealing with machine and process values. These could be anything,
                # but typically only expect either a string or single value list.
                # Warn about anything unexpected and grab the value anyway.
                if isinstance(values, str) or (
                    isinstance(values, list) and len(values) == 1
                ):
                    # What we expected.
                    settings[key] = values

                else:
                    print(
                        f"Warning: Unexpected type for machine/process override."
                        f" This is probably OK, but should be checked to make sure."
                        f"\n  In override `{metadata.name}` for setting `{key}`:"
                        f"\n  Expected either a single value list or a string."
                        f"\n  Got (Values = {values})"
                    )
                    # Not what we expected, but still probably fine.
                    settings[key] = values


class ProjectDiffSet:
    """Group of project/3mf settings differences/comparisons.

    All differences/comparisons in a ProjectDiffSet instance are relative to a reference
    PresetGroup specified when the instance is initialised. The default reference
    PresetGroup is "system", which follows the BambuStudio model for project configs
    (as implemented by "different_settings_to_system" in the project_settings.config
    file).

    ProjectDiffSet uses the following information to generate differences:
    - Setting values for the project presets.
    - Setting values for the reference presets that project presets inherit from (may
    be unset).
    - If the the project preset is an override, the setting values the project preset
    inherits from its parent (may be unset).

    The ProjectDiffSet presents the following data when differences exist:
    - The setting value and the reference value (may be unset).
    - A reference value flag of REFERENCE, COMPLEX, SPECIAL or UNSET.
    - A settings value flag of DIFFERENT, COMPLEX, SPECIAL or OVERRIDE.
    - OVERRIDE represents values that Bambu Studio highlights in orange.
    A more long winded explanation (because I need it):
            - The project preset is created by inheriting a user or project preset, but
            overrides the value inherited from its parent (orange in BS GUI).
            - The parent value is different to the reference value (for a regular
            difference, this would be a black value in the BS GUI).
            - The system reference value from the system preset is different to both
            of the above values.
    The OVERRIDE parent values are not recorded.
    - COMPLEX means that type of the setting is too complicated for this difference
    tool to evaluate (e.g. a list of lists or other similar nested or mixed type
    objects.)
    - SPECIAL means that these values require special handling or may have been hacked
    to deliver package functionality - refer to the package documentation on the best
    way to handle these setting/value pairs.

    Note: this class is intended to allow comparison of multiple settings in
    multiple files. A separate class should be implemented to compare two items (e.g.
    filaments). I haven't implemented this for now, as this is largely available in the
    BS GUI (albeit with no export function).
    """

    _reference_group: PresetGroup
    # Data is a dict of sparse matrices storing values and index info.
    _data: dict[PresetType, DiffMatrix]
    # The current target row for writing out Excel data.
    _write_row: int = 1

    def __init__(self, reference_group: PresetGroup = PresetGroup.SYSTEM) -> None:
        """Create preset diff."""
        # The default preset group is SYSTEM, which more or less follows the BS
        # representation.
        self._reference_group = reference_group
        self._data = {}
        for preset_type in PresetType:
            self._data[preset_type] = DiffMatrix()

    @staticmethod
    def _extract_str(value: SettingValue, diff_type: DiffType) -> tuple[str, DiffType]:
        """Very simple type checking and string extraction.

        Returns (simple string value or repr of complex value, diff_type if string value
        found | DiffType.Complex if complex value)

        Value is simple if it is a string or list with one string member.
        """
        if isinstance(value, str):
            return value, diff_type
        if isinstance(value, list) and len(value) == 1 and isinstance(value[0], str):
            return value[0], diff_type

        return repr(value), DiffType.COMPLEX

    def _add_diff(
        self, settings: AllNodeSettings, inherited: AllNodeSettings | None = None
    ) -> None:
        """Add any differences in preset to the DiffSet."""
        # Easy checks first.

        # Validate that this is a unique project preset.
        metadata = settings.metadata
        if self._data[metadata.preset_type].column_exits(metadata):
            raise IndexError(f"Repeated difference for preset '{metadata}'")

        # Terminate if there is no reference data (should not happen).
        if settings.ref_metadata is None:
            raise ValueError(
                f"Preset '{metadata}' is missing difference reference data"
                f"\nThis should not be possible!"
            )

        # Comparison process, which is bit messy. I think the following truth table
        # covers it:
        # v1 != v2 != v3
        # Preset  Ref.    Over.   Over.   Preset   Ref.
        # Value   Value   type?   Value   Flag     Flag  Record?  Notes
        #   v1      v1     Any      NA    NO_DIFF  REF   No       Shouldn't happen,
        #                                                         check anyway.
        #   v1      Unset  No       NA    DIFF     UNSET Yes
        #   v1      Unset  Yes      Unset DIFF     UNSET Yes
        #   v1      Unset  Yes      v1    DIFF     UNSET Yes
        #   v1      Unset  Yes      v3    OVER.    UNSET Yes
        #
        #   v1      v2     No       NA    DIFF     REF   Yes
        #   v1      v2     Yes      Unset DIFF     REF   Yes
        #   v1      v2     Yes      v1    DIFF     REF   Yes
        #   v1      v2     Yes      v2    DIFF     REF   Yes      Shouldn't happen!
        #   v1      v2     Yes      v3    OVER.    REF   Yes
        #
        # To really do this properly, we should cross check that value types match,
        # which will get ugly for nest types (probably needs pydantic), and then
        # for differences in the values themselves.
        # HOWEVER, that's a bit much work for the simple difference tool that I'm aiming
        # for in the first pass. So for the first pass, we take the following shortcuts.
        # A) Assume the inheritance tree is definitive for differences. So we take it
        # on faith the differences exist if the tree says so and don't check.
        # B) We do check that value types are either a string or a single string value
        # in a list, which we will need to do make sure override types match the
        # presets they inherit from (see implementation).
        # C) Anything more complex than (B) is recorded as TOO_COMPLEX for the
        # simple difference. Up to the user to resolve. (For most cases this probably
        # won't be an issue).
        # Revisit this if complex types happen more often than I expect.

        for setting_name, value in settings.source_subtree.items():
            if setting_name in settings.reference_subtree:
                ref_value = settings.reference_subtree[setting_name]
                ref_value, ref_type = self._extract_str(ref_value, DiffType.REFERENCE)

            else:
                ref_value = ""
                ref_type = DiffType.UNSET

            value, value_type = self._extract_str(value, DiffType.DIFFERENCE)

            if setting_name in SPECIAL_TYPES:
                # Special type, don't bother checking override details, just update
                # types.
                value_type = DiffType.SPECIAL
                ref_type = DiffType.SPECIAL

            elif (
                value_type == DiffType.DIFFERENCE
                and inherited
                and setting_name in inherited.source_subtree
            ):
                # Handle the odd duck if we have a difference.
                # No point checking overrides for the complex case.
                inherit_value = inherited.source_subtree[setting_name]
                inherit_value, inherit_type = self._extract_str(
                    inherit_value, DiffType.DIFFERENCE
                )
                # now a long string of checks to see if we have a meaningful
                # override.
                if inherit_type == DiffType.DIFFERENCE and inherit_value != value:
                    if ref_type == DiffType.UNSET or (
                        ref_type == DiffType.REFERENCE and ref_value != inherit_value
                    ):
                        # value != inherit != (unset | ref_value).
                        value_type = DiffType.OVERRIDE

            # Record value, add setting name to row indices (likely to be duplicate
            # calls, not worth chasing efficiency here).
            self._data[settings.metadata.preset_type].add_value(
                setting_name, settings.metadata, value, value_type
            )
            self._data[settings.metadata.preset_type].add_value(
                setting_name, settings.ref_metadata, ref_value, ref_type
            )

    def add_project_presets(self, project: ThreeMFPresets) -> None:
        """Create diffs for all project presets in a 3mf file."""
        for settings in project.presets.project_presets(self._reference_group):

            # If settings are for an OVERRIDE preset, then the diff set needs to
            # know the parent settings for preset the that the override preset is built
            # from. Gather them here if needed.
            inherited: AllNodeSettings | None = None
            if settings.metadata.override_inherits:
                inherited = project.presets.all_node_settings(
                    settings.metadata.preset_type,
                    node_name=settings.metadata.override_inherits,
                    ref_group=self._reference_group,
                )
                # Previously tried to do a basic check for consistency of reference
                # values, but this falls down if trying to compare anything other
                # than a system ref group, so I've dropped it and just take it on
                # faith that BS is doing the right thing.

            self._add_diff(settings, inherited)

    @staticmethod
    def _xlsx_cell(  # type: ignore[no-untyped-def]
        ws, row: int, column: int, value: str, cell_format: DiffType | CellFormat
    ) -> int:
        """Set the value and format for an xlsx cell.

        returns row + 1 (useful if moving down after each cell).
        """
        cell = ws.cell(row=row, column=column, value=value)
        # can add Neutral if I want a special value.
        if isinstance(cell_format, DiffType):
            match cell_format:
                case DiffType.DIFFERENCE:
                    cell.style = CellFormat.NORMAL
                case DiffType.REFERENCE:
                    cell.style = CellFormat.GOOD
                case DiffType.UNSET:
                    cell.style = CellFormat.BAD
                case DiffType.OVERRIDE:
                    cell.style = CellFormat.INPUT
                case DiffType.COMPLEX:
                    cell.style = CellFormat.NOTE
                case DiffType.SPECIAL:
                    cell.style = CellFormat.NEUTRAL
                case _:
                    cell.style = "Normal"
        else:
            cell.style = cell_format

        return row + 1

    def _xlsx_empty_line(self) -> None:
        """Add an empty line to current sheet."""
        self._write_row += 1

    def _xlsx_legend(self, ws) -> None:  # type: ignore[no-untyped-def]
        """Write legend information into worksheet."""
        self._write_row = self._xlsx_cell(
            ws,
            row=self._write_row,
            column=1,
            value="Normal/difference value",
            cell_format=DiffType.DIFFERENCE,
        )
        self._write_row = self._xlsx_cell(
            ws,
            row=self._write_row,
            column=1,
            value="Reference value",
            cell_format=DiffType.REFERENCE,
        )
        self._write_row = self._xlsx_cell(
            ws,
            row=self._write_row,
            column=1,
            value="Unset (problem) value",
            cell_format=DiffType.UNSET,
        )
        self._write_row = self._xlsx_cell(
            ws,
            row=self._write_row,
            column=1,
            value="Value too complex to difference",
            cell_format=DiffType.COMPLEX,
        )
        self._write_row = self._xlsx_cell(
            ws,
            row=self._write_row,
            column=1,
            value="Override value",
            cell_format=DiffType.OVERRIDE,
        )
        self._write_row = self._xlsx_cell(
            ws,
            row=self._write_row,
            column=1,
            value=(
                "Special. Value type needs special/alternative handling. See package"
                " documentation for details."
            ),
            cell_format=DiffType.SPECIAL,
        )

    def _xls_diff_table(  # type: ignore[no-untyped-def]
        self, ws, preset: PresetType
    ) -> None:
        """Write diff table to ws."""
        self._write_row = self._xlsx_cell(
            ws, self._write_row, 1, preset.value, CellFormat.HEADING4
        )

        data = self._data[preset]
        for cell in data.table_cells():
            self._xlsx_cell(
                ws, self._write_row + cell.row, cell.column + 1, cell.value, cell.format
            )

        self._write_row += data.row_count()

    def save_xlsx(self, xlsx_file: Path | None = None) -> None:
        """Write the DiffSet to xlsx file using openpyxl."""
        if not xlsx_file:
            xlsx_file = Path().cwd() / "Preset diffs.xlsx"

        if xlsx_file.exists():
            c = input("Difference file exists. y/Y to overwrite> ")

            if c[0] != "y" and c[0] != "Y":
                print("Differences not saved.")
                return

        # Create workbook and worksheets and populate. We could also modify this to dump
        # the tables to a single worksheet, but I think the per worksheet is better for
        # separating machine/process/filament.
        wb = Workbook()
        for preset in PresetType:
            ws = wb.create_sheet(title=preset.value)
            self._write_row = 1
            self._xlsx_legend(ws)
            self._xlsx_empty_line()
            self._xls_diff_table(ws, preset)

        # Clean up - delete empty ws[0]
        wb.remove(wb.worksheets[0])
        wb.save(filename=str(xlsx_file.resolve()))


def presets_to_xlsx(cl_args: Namespace) -> None:
    """Create differences from arguments."""
    if cl_args.source:
        working = Path(cl_args.source).resolve()
    else:
        working = Path.cwd()

    # For non-default system or user presets, create a StandardPreset instance and use
    # it when creating the first instance of ThreeMFConfig (it will be ignored on
    # second and later calls).
    #  standard_presets = StandardPresets(appdata_path, user_id)
    #  this_config = ThreeMFConfig(...., standard_presets=standard_presets)
    # Could add this to command line arguments if there is a real call for it.
    # For now, I'm assuming that the vast majority of users won't need this.

    # Create a diff set that mimics the standard BambuStudio diff to system.
    diffs = ProjectDiffSet(cl_args.ref)

    if working.is_file():
        this_3mf = ThreeMFPresets(working)
        diffs.add_project_presets(this_3mf)
        working = working.parent
    else:
        for filepath in working.iterdir():
            if filepath.is_file() and filepath.suffix.lower() == THREE_MF_EXT:
                # Grab metadata, discard after generating differences.
                this_3mf = ThreeMFPresets(filepath)

                # This should have loaded all key configuration data.
                # Generate differences from system settings.
                diffs.add_project_presets(this_3mf)

    if cast(str, cl_args.output)[-5:] != ".xlsx":
        xlsx_path = working / (cl_args.output + ".xlsx")
    else:
        xlsx_path = working / cl_args.output

    diffs.save_xlsx(xlsx_path)


def dump_json(cl_args: Namespace) -> None:
    """Export settings from preset(s) in .3mf file."""
    filepath = Path(cl_args.source).resolve()
    this_3mf = ThreeMFPresets(filepath)

    selection = choose(
        options=[preset_type.value for preset_type in PresetType],
        header=(
            "Select the preset type you want to export."
        ),
        allow_multi=False,
        no_action="Return without exporting.",
    )

    if not selection:
        return

    preset_type = PresetType(selection[0])
    names = this_3mf.presets.preset_names(preset_type)

    selection = choose(
        names,
        header=(
            "\nEnter one or more comma separated numbers to choose the presets"
            "\nthat you want to export. E.g. 1, 3, 5<Enter>. Filenames will be"
            "\nthe preset name with '.json' suffix. Refer package documentation"
            "\nfor explanation of override naming (random letters on end of names)."
        ),
        allow_multi=True,
        no_action="Return without exporting.",
    )

    if not selection:
        return

    working = filepath.parent
    for name in selection:
        settings = this_3mf.presets.all_node_settings(preset_type, name)
        filepath = working / (name + ".json")

        with open(filepath, "w", encoding=DEFAULT_ENCODING) as fp:
            json.dump(settings.source_subtree, fp)


def get_parser() -> ArgumentParser:
    """Create argument parser for command line."""
    this_parser = ArgumentParser(
        description="Simple difference/dumper utility for BambuStudio .3mf files."
    )

    subparsers = this_parser.add_subparsers(
        help="<command> --help, -h for subcommand help.", required=True
    )

    parser_diff = subparsers.add_parser(
        "diff",
        description="Generate differences from file or folder.",
    )

    parser_diff.add_argument(
        "--source",
        help=(
            "Specify a source .3mf file or folder containing .3mf files."
            " Defaults to cwd()."
        ),
    )

    parser_diff.add_argument(
        "--output",
        help="Filename for output without suffix (.xlsx or .json files will"
        " be created based on mode). Defaults to `preset_diffs`",
        default="preset_diffs",
    )

    # Ended up deciding that I don't trust anything other than system, so removed
    # this capability altogether.
    # parser_diff.add_argument(
    #    "--ref",
    #    help=(
    #        "Name of the reference group for differencing"
    #        " (`system` | `user` | `project`)."
    #        " Defaults to `system`. 'system' is robust, others are untested and should"
    #        " be used with extreme caution (I recommend only using system)."
    #    ),
    #    type=PresetGroup,
    #    default=PresetGroup.SYSTEM,
    # )
    parser_diff.set_defaults(ref=PresetGroup.SYSTEM)

    parser_diff.set_defaults(func=presets_to_xlsx)

    parser_dump = subparsers.add_parser(
        "export",
        description=(
            "Export json for one or more presets from a file. An interactive menu"
            " system will guide you through selecting preset type and preset names"
            " for export."
        ),
    )

    parser_dump.add_argument(
        "source",
        help=("Specify a source .3mf file."),
    )

    parser_dump.set_defaults(func=dump_json)

    return this_parser


if __name__ == "__main__":
    parser = get_parser()
    args = parser.parse_args()
    args.func(args)



#!/usr/bin/env python
# cspell:ignore Bambu
"""Common supporting classes and constants for slicer tools."""

from enum import StrEnum
from typing import NamedTuple, Sequence, Any

# create some type aliases
type SettingValue = str | list[str]
type SettingsDict = dict[str, SettingValue]

# Navigation and file name components
DEFAULT_ENCODING = "utf-8"

# json key elements
FROM = "from"
INHERITS = "inherits"
NAME = "name"


class PresetType(StrEnum):
    """Preset types used by Bambu Studio."""

    FILAMENT = "filament"
    MACHINE = "machine"
    PROCESS = "process"


class PresetGroup(StrEnum):
    """Preset groups used by Bambu Studio.

    System refers to presets from the system folder.
    User refers to presets from the user folder.
    Project refers to a filament, process or machine preset in a 3mf file.
    Override refers to the project settings from a 3mf file that override any of
    of the above (this is strictly also project preset, but as it is handled somewhat
    differently to the other presets by BBS, I've created a special category).
    """

    OVERRIDE = "override"
    PROJECT = "project"
    USER = "user"
    SYSTEM = "system"


class NodeMetadata(NamedTuple):
    """Summary metadata relating to Preset Nodes.

    Used for initialising PresetNodes and in generating complete settings lists from
    tree walk.
    """

    # Unique name for preset in project file.
    name: str
    # Source filename for the preset. Source file may contain multiple presets.
    filename: str
    group: PresetGroup
    preset_type: PresetType

    # The override_inherits field deals with an irritating quirk of project
    # overrides - they are defined as differences to a system preset. This system
    # preset is specified as the "inherits_group" value for the project override (i.e.
    # the parent for the differences is relative to the system preset named in
    # "inherits_group").
    #
    # However, in the Bambu Studio interface, orange highlighted values show differences
    # relative to the project specified immediate parent of the override, which may not
    # be the same as the system preset above! The immediate parent is named in:
    # "<printer|print|filament>_settings_id". (i.e. the project settings includes two
    # inheritance paths for overrides!)
    #
    # To capture this information in this package, we need to know both the system
    # preset and the immediate parent preset.
    #
    # I could have handled this by introducing a dummy key/value pair into the
    # the settings dictionary, but this would probably cause bigger problems with
    # end users trying to create a preset containing this dummy parameter.
    #
    # So instead we do a special here to handle the unicorn sparkly tears case
    # of project settings.
    #
    # override_inherits is the name of the immediate parent preset for the override. It
    # may be different from the "inherits" value used to generate
    # "differences_to_system" values.
    #
    # Defaults to empty string and should only be set for group = Preset.OVERRIDE.
    override_inherits: str = ""


class AllNodeSettings(NamedTuple):
    """Container for all of the settings for a preset node (from tree roll up)."""

    # metadata for the source/target node
    metadata: NodeMetadata
    # metadata for the reference node if defined.
    ref_metadata: NodeMetadata | None
    # Settings generated by walking the inheritance tree from the source/target node
    # up to but not including the reference node if defined, or all settings if
    # reference node is not defined).
    # In effect, these are the source node settings that are different to the reference
    # node settings.
    # (If reference settings exist, these settings either override or augment the
    # reference settings.)
    source_subtree: SettingsDict
    # Settings generated by walking the inheritance tree from the reference node to
    # the root of the inheritance tree.
    reference_subtree: SettingsDict


def choose(
    options: Sequence[str | tuple[str, Any]],
    header: str | None = None,
    allow_multi: bool = False,
    no_action: str | None = None,
    help_text: str | None = None,
) -> list[Any] | None:
    """Return user selection or multi-selection from a list of options.

    Arguments:
        options {Sequence[Union[str, Tuple[str, Any]]]} -- A list of options, where
            each may be either:
            * a string, where string is returned if the option is selected; or
            * a of tuple (option_description, return_value), where the return_value
              is returned if the option is selected.

    Keyword Arguments:
        header {str} -- Header/description text for the selection (default: {None})
        allow_multi {bool} -- Allow multi-selection if True (default: {False})
        no_action {str} -- Description value for no selection. If this argument is not
            set or is None, the user must select from the options list.
            (default: {None})
        help_text {str} -- Detailed help text. (default: {None})

    Returns:
        Optional[List[Any]] -- List of return values. If the allow_multi is False, the
        list will contain only one element. If the user has selected no_action, the
        return value is None.

    To select multiple options, enter a comma separated list of values.

    """
    opt_tuple_list = list()
    for this_opt in options:
        if isinstance(this_opt, tuple):
            opt_tuple_list.append(this_opt)
        else:
            opt_tuple_list.append((this_opt, this_opt))

    while True:
        print()
        if header is not None:
            print(header)
            print()

        for i, (key, value) in enumerate(opt_tuple_list):
            print(f"{i + 1: 3d}) {key}")
        if no_action is not None:
            print("  0) " + no_action)
        if help_text is not None:
            print("  h) Help.")

        values: Any = ""
        try:
            # splits on commas, strips white space from values, converts to int
            print()
            values = input("Choose> ").strip()
            values = list(map(lambda x: int(x.strip()) - 1, values.split(",")))

            if no_action is not None and -1 in values:
                return None

            values = [
                opt_tuple_list[value][1]
                for value in values
                if 0 <= value < len(opt_tuple_list)
            ]

            if not values:
                continue

            if allow_multi:
                return values

            if len(values) == 1:
                return values[0:1]

        except (ValueError, IndexError):
            if help_text is not None and values == "h":
                print("-" * 80)
                print()
                print(help_text)
                print()
                print("-" * 80)
                continue
            else:
                print("Invalid input, please try again")
